package nz.co.sisu.tools.jetty.util;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ClasspathBuilder {

    private static final Logger LOG = LoggerFactory.getLogger(ClasspathBuilder.class);

    private static final String SEP = ",";

    public static String appendPaths(String[] parts) {
        String result = toPath(parts, SEP);
        LOG.debug("appendPaths result: {}", result);
        return result;
    }

    public static URL[] appendUrls(String[] parts) {
        //
        // FIXME ... this is somewhat ugly, we re-split a string created earlier
        //
        URL[] result = null;
        if (parts != null) {
            List working = new ArrayList();
            for (int i = 0; i < parts.length; i++) {
                String[] items = parts[i].split(SEP);
                for (int k = 0; k < items.length; k++) {
                    String item = items[k];
                    try {
                        working.add(new File(item).toURL());
                    }
                    catch (MalformedURLException e) {
                        LOG.warn("File to URL conversion problem: " + item, e);
                    }
                }
            }
            result = (URL[]) working.toArray(new URL[working.size()]);

            if (LOG.isDebugEnabled()) {
                LOG.debug("appendPaths result: {}", toPath(parts, SEP));
            }
        }
        return result;
    }

    public static String buildPath(String base, String[] include, String[] exclude) {
        FileSystemVisitor v = new FileSystemVisitor(include, exclude);
        v.traverse(toFile(base));
        String[] files = v.getAcceptedAsNormalisedFilenames();
        String result = toPath(files, SEP);
        LOG.debug("buildPath result: {}", result);
        return result;
    }

    private static File toFile(String path) {
        File f = new File(path);
        if (!f.exists()) {
            throw new IllegalArgumentException("Invalid path [" + path + "]");
        }
        return f;
    }

    private static String toPath(String[] parts, String sep) {
        StringBuffer b = new StringBuffer();
        if (parts != null) {
            for (int i = 0; i < parts.length; i++) {
                if (i > 0) {
                    b.append(sep);
                }
                b.append(parts[i]);
            }
        }
        return b.toString();
    }

    private static class FileSystemVisitor {

        private static final Comparator FILE_COMPARATOR = new FileComparator();

        private String[] includes;
        private String[] excludes;
        private Set filePaths;

        public FileSystemVisitor(String[] includes, String[] excludes) {
            this.includes = includes;
            this.excludes = excludes;
            this.filePaths = new LinkedHashSet();
        }

        public void traverse(File base) {
            visit(base);
        }

        public String[] getAcceptedAsNormalisedFilenames() {
            Set results = new LinkedHashSet();
            Iterator iter = getFilePaths().iterator();
            while (iter.hasNext()) {
                File f = (File) iter.next();
                results.add(normaliseFilename(f.getPath()));
            }
            return (String[]) results.toArray(new String[results.size()]);
        }

        private void visit(File f) {
            if (f.exists()) {
                if (acceptDirectory(f)) {
                    visitDirectory(f);
                }
                else if (acceptFile(f)) {
                    visitFile(f);
                }
                else {
                    LOG.info("Not file or directory: {}", f);
                }
            }
            else {
                LOG.info("Not found: {}", f);
            }
        }

        private void visitFile(File f) {
            if (!acceptFile(f)) {
                throw new IllegalStateException("Not a file: " + f);
            }

            if (acceptFilename(f)) {
                getFilePaths().add(f);
            }
        }

        private void visitDirectory(File d) {
            if (!acceptDirectory(d)) {
                throw new IllegalStateException("Not a directory: " + d);
            }
            File[] files = d.listFiles();
            Arrays.sort(files, FILE_COMPARATOR);

            visitFiles(files);
            visitDirectories(files);
        }

        private void visitDirectories(File[] f) {
            for (int i = 0; i < f.length; i++) {
                if (acceptDirectory(f[i])) {
                    visitDirectory(f[i]);
                }
            }
        }

        private void visitFiles(File[] f) {
            for (int i = 0; i < f.length; i++) {
                if (acceptFile(f[i])) {
                    visitFile(f[i]);
                }
            }
        }

        private boolean acceptFile(File f) {
            return f.isFile() && !f.isHidden();
        }

        private boolean acceptDirectory(File f) {
            return f.isDirectory() && !f.isHidden();
        }

        private boolean acceptFilename(File f) {
            boolean result = true;
            String path = normaliseFilename(f.getPath());

            String[] includes = getIncludes();
            if (includes != null) {
                result = false;
                for (int i = 0; i < includes.length; i++) {
                    if (path.matches(includes[i])) {
                        result = true;
                        break;
                    }
                }
                if (!result) {
                    LOG.debug("No include match, path[{}]", path);
                }
            }

            if (result) {
                String[] excludes = getExcludes();
                if (excludes != null) {
                    for (int i = 0; i < excludes.length; i++) {
                        if (path.matches(excludes[i])) {
                            result = false;
                            LOG.debug("Exclude match: path[{}], exclude[{}]", path, excludes[i]);
                            break;
                        }
                    }
                }
            }

            return result;
        }

        private String normaliseFilename(String filename) {
            String result = null;
            if (filename != null) {
                result = filename.replace('\\', '/');
            }
            return result;
        }

        private String[] getIncludes() {
            return includes;
        }

        private String[] getExcludes() {
            return excludes;
        }

        private Set getFilePaths() {
            return filePaths;
        }

    }

    private static class FileComparator implements Comparator {
        public int compare(Object o1, Object o2) {
            File f1 = (File) o1;
            File f2 = (File) o2;

            return f1.getPath().toLowerCase().compareTo(f2.getPath().toLowerCase());
        }
    }

}
