package nz.co.sisu.tools.jetty.util;

import java.net.URL;
import java.net.URLClassLoader;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ServicesClassLoader extends URLClassLoader {

    private static final Logger LOG = LoggerFactory.getLogger(ServicesClassLoader.class);

    private static final Set STANDARD_CLASSNAME_PREFIXES;

    public ServicesClassLoader(URL[] urls, ClassLoader parent) {
        super(urls, determineParent(parent));
    }

    public ServicesClassLoader(URL[] urls) {
        this(urls, null);
    }

    /**
     * Same parent class loader determination as in
     * {@link org.mortbay.jetty.webapp.WebAppClassLoader}
     */
    private static ClassLoader determineParent(ClassLoader parent) {
        ClassLoader result = parent;
        String location = "supplied";
        if (result == null) {
            result = Thread.currentThread().getContextClassLoader();
            location = "context";
        }
        if (result == null) {
            result = ServicesClassLoader.class.getClassLoader();
            location = "current";
        }
        if (result == null) {
            result = ClassLoader.getSystemClassLoader();
            location = "system";
        }
        if (result == null) {
            location = "none";
        }
        LOG.debug("Parent class loader: " + location);
        return result;
    }

    /**
     * Non-standard resource loading, like in the servlet spec, attempts to load
     * the resource locally then delegates to parent.
     */
    public URL getResource(String name) {
        URL url = findResource(name);
        if (url == null) {
            ClassLoader parent = getParent();
            if (parent != null) {
                url = parent.getResource(name);
                if (url != null) {
                    LOG.debug("getResource(" + name + ") [parent] " + url);
                }
                else {
                    LOG.debug("getResource(" + name + ") [not found]");
                }
            }
        }
        else {
            LOG.debug("getResource(" + name + ") [local] " + url);
        }
        return url;
    }

    /**
     * Non-standard class loading, like in the servlet spec, attempts to load
     * the resource locally then delegates to parent.
     */
    protected synchronized Class loadClass(String name, boolean resolve) throws ClassNotFoundException {
        // First, check if the class has already been loaded
        Class c = findLoadedClass(name);
        if (c == null) {
            if (isStandardClass(name)) {
                // parent then local

                ClassLoader parent = getParent();
                if (parent != null) {
                    try {
                        c = parent.loadClass(name);
                        LOG.debug("loadClass(" + name + "," + resolve + ") parent");
                    }
                    catch (ClassNotFoundException e) {
                        c = findClass(name);
                        LOG.debug("loadClass(" + name + "," + resolve + ") local");
                    }
                }
                else {
                    c = findClass(name);
                    LOG.debug("loadClass(" + name + "," + resolve + ") local");
                }
            }
            else {
                // local then parent
                try {
                    c = findClass(name);
                    LOG.debug("loadClass(" + name + "," + resolve + ") local");
                }
                catch (ClassNotFoundException e) {
                    ClassLoader parent = getParent();
                    if (parent != null) {
                        c = parent.loadClass(name);
                        LOG.debug("loadClass(" + name + "," + resolve + ") parent");
                    }
                    else {
                        throw new ClassNotFoundException(name);
                    }
                }
            }
        }
        else {
            LOG.debug("loadClass(" + name + "," + resolve + ") loaded");
        }
        if (resolve) {
            resolveClass(c);
        }
        return c;
    }

    private boolean isStandardClass(String name) {
        boolean result = false;
        Iterator iter = STANDARD_CLASSNAME_PREFIXES.iterator();
        while (iter.hasNext()) {
            String prefix = (String) iter.next();
            if (name.startsWith(prefix)) {
                result = true;
                break;
            }
        }
        return result;
    }

    static {
        Set standardClassnamePrefixes = new HashSet();
        standardClassnamePrefixes.add("java.");
        standardClassnamePrefixes.add("javax.");
        standardClassnamePrefixes.add("org.w3c.dom.");
        standardClassnamePrefixes.add("org.xml.sax.");

        STANDARD_CLASSNAME_PREFIXES = Collections.unmodifiableSet(standardClassnamePrefixes);
    }

}
